/* --------------------------------------------------------------------------
  FILE        : device.c 				                             	 	        
  PURPOSE     : Device (chip) specific initialization function
  PROJECT     : DA8xx ROM boot loader
  AUTHOR      : Daniel Allred
  DESC        : This file provides low-level init functions for use
                  in the UBL for booting an application.
-------------------------------------------------------------------------- */ 

// General type include
#include "tistdtypes.h"

// This module's header file
#include "device.h"

// Debug module
#include "debug.h"

// Utility functions
#include "util.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/


/************************************************************
* Local Variable Definitions                                *
\***********************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/


/************************************************************
* Global Function Definitions                               *
************************************************************/
Uint32 DEVICE_init()
{
  Uint32 status = E_PASS;

#ifndef SKIP_LOW_LEVEL_INIT
  // System PSC Setup
  DEVICE_PSCInit();

  // System PLL Setup
  if (status == E_PASS) status |= DEVICE_PLL0Init(DEVICE_PLL0_MULTIPLIER);
    
  // DDR2 Timing Setup
  if (status == E_PASS) status |= DEVICE_SDRAMInit();
#endif

#ifdef UBL_SPI
  if (status == E_PASS) status |= DEVICE_SPIInit();
#endif	

  // AEMIF Setup
  if (status == E_PASS) status |= DEVICE_EMIFInit();

  return status;
}


void DEVICE_LPSCTransition(Uint8 pscnum, Uint8 module, Uint8 domain, Uint8 state)
{
  DEVICE_PSCRegs *PSC;
  
  if (pscnum == 0)
    PSC = PSC0;
  else if(pscnum == 1)
    PSC = PSC1;
  else
    return; 

  // Wait for any outstanding transition to complete
  while ( (PSC->PTSTAT) & (0x00000001 << domain) );
  
  // If we are already in that state, just return
  if (((PSC->MDSTAT[module]) & 0x1F) == state) return;
    
  // Perform transition
  PSC->MDCTL[module] = ((PSC->MDCTL[module]) & (0xFFFFFFE0)) | (state);
  PSC->PTCMD |= (0x00000001 << domain);

  // Wait for transition to complete
  while ( (PSC->PTSTAT) & (0x00000001 << domain) );
  
  // Wait and verify the state
  while (((PSC->MDSTAT[module]) & 0x1F) != state);	
}

void DEVICE_PSCInit()
{                                                                    
  Uint32 i;
  
  // PSC0, domain 0 init
  while ((PSC0->PTSTAT) & 0x00000001);

  for( i = 3 ; i <= 4 ; i++ )
	if((PSC0->MDCTL[i] & 0x1F) != PSC_ENABLE)
      PSC0->MDCTL[i] = (PSC0->MDCTL[i] & 0xFFFFFFE0) | PSC_ENABLE;

  if((PSC0->MDCTL[7] & 0x1F) != PSC_ENABLE)
      PSC0->MDCTL[7] = (PSC0->MDCTL[7] & 0xFFFFFFE0) | PSC_ENABLE;

  for( i =  9; i <= 12 ; i++ )
    if((PSC0->MDCTL[i] & 0x1F) != PSC_ENABLE)
      PSC0->MDCTL[i] = (PSC0->MDCTL[i] & 0xFFFFFFE0) | PSC_ENABLE;
    
  // Do Always-On Power Domain Transitions
  PSC0->PTCMD |= 0x00000001;
  while ((PSC0->PTSTAT) & 0x00000001);
  
  // PSC1, domain 1 init
  while ((PSC1->PTSTAT) & 0x00000001);

  if((PSC1->MDCTL[3] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[3] = (PSC1->MDCTL[3] & 0xFFFFFFE0) | PSC_ENABLE;

  if((PSC1->MDCTL[6] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[6] = (PSC1->MDCTL[6] & 0xFFFFFFE0) | PSC_ENABLE;

  for( i = 12 ; i <= 13 ; i++ )
    if((PSC1->MDCTL[i] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[i] = (PSC1->MDCTL[i] & 0xFFFFFFE0) | PSC_ENABLE;

  for( i = 22 ; i <= 23 ; i++ )
    if((PSC1->MDCTL[i] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[i] = (PSC1->MDCTL[i] & 0xFFFFFFE0) | PSC_ENABLE;

  if((PSC1->MDCTL[26] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[26] = (PSC1->MDCTL[26] & 0xFFFFFFE0) | PSC_ENABLE;

  if((PSC1->MDCTL[31] & 0x1F) != PSC_ENABLE)
      PSC1->MDCTL[31] = (PSC1->MDCTL[31] & 0xFFFFFFE0) | PSC_ENABLE;
    
  // Do Always-On Power Domain Transitions
  PSC1->PTCMD |= 0x00000001;
  while ((PSC1->PTSTAT) & 0x00000001);
}

void DEVICE_pinmuxControl(Uint32 regOffset, Uint32 mask, Uint32 value)
{
  SYSTEM->KICKR[0] = 0x83e70b13;  // Kick0 register + data (unlock)
  SYSTEM->KICKR[1] = 0x95a4f1e0;  // Kick1 register + data (unlock)

  SYSTEM->PINMUX[regOffset] &= ~mask;
  SYSTEM->PINMUX[regOffset] |= (mask & value);

  SYSTEM->KICKR[0] = 0x00000000;  // Kick0 register + data (lock)
  SYSTEM->KICKR[1] = 0x00000000;  // Kick1 register + data (lock)
}

static const DEVICE_BootMode bootModes[] = {
	DEVICE_BOOTMODE_I2C0_MASTER,
	DEVICE_BOOTMODE_I2C0_SLAVE,
	DEVICE_BOOTMODE_NOR_EMIFA, DEVICE_BOOTMODE_NOR_EMIFA,
	DEVICE_BOOTMODE_UHPI, DEVICE_BOOTMODE_UHPI,
	DEVICE_BOOTMODE_I2C1_MASTER,
	DEVICE_BOOTMODE_I2C1_SLAVE,
	DEVICE_BOOTMODE_SPI0_EEPROM,
	DEVICE_BOOTMODE_SPI1_EEPROM,
	DEVICE_BOOTMODE_SPI0_FLASH, DEVICE_BOOTMODE_SPI0_FLASH,
	DEVICE_BOOTMODE_SPI1_FLASH, DEVICE_BOOTMODE_SPI1_FLASH,
	DEVICE_BOOTMODE_NAND_EMIFA_8BIT, DEVICE_BOOTMODE_NAND_EMIFA_8BIT,
	DEVICE_BOOTMODE_NAND_EMIFA_16BIT,
	DEVICE_BOOTMODE_NONE,
	DEVICE_BOOTMODE_SPI0_SLAVE,
	DEVICE_BOOTMODE_SPI1_SLAVE,
	DEVICE_BOOTMODE_UART2,
	DEVICE_BOOTMODE_THB,
	DEVICE_BOOTMODE_UART0,
	DEVICE_BOOTMODE_UART1,
	DEVICE_BOOTMODE_ESF, DEVICE_BOOTMODE_ESF,
	DEVICE_BOOTMODE_USB11,
	DEVICE_BOOTMODE_USB20,
	DEVICE_BOOTMODE_MMC,
	DEVICE_BOOTMODE_RMII,
	DEVICE_BOOTMODE_EMU_DEBUG,
	DEVICE_BOOTMODE_NONE,
};


DEVICE_BootMode DEVICE_bootMode( void )
{
	Uint32 i;
	// Read BOOT pins and set bootMode
	// Using a look-up table
	i = SYSTEM->BOOTCFG & 0xFFFF;
	// Bit sequence used for LUT is 7,2,1,0,3
	// Translate: 76543210 to ___72103
	// (__210 << 1) + (7___3___ >> 3) = _210_ + 7___3 = 72103
	i = ((i & 0x0007) << 1) + ((i & 0x0088) >> 3);			// _210_ + 7___3

	return bootModes[i];
}

DEVICE_BusWidth DEVICE_emifBusWidth( void )
{
  if (DEVICE_bootMode() == DEVICE_BOOTMODE_NAND_EMIFA_8BIT)
  {
    return DEVICE_BUSWIDTH_8BIT;
  }
  else
  {
    return DEVICE_BUSWIDTH_16BIT;
  }
}

Uint32 DEVICE_EMIFInit()
{

  DEVICE_pinmuxControl(13,0xFFFFFFFF,0x11111111);
  DEVICE_pinmuxControl(14,0xFFFFFFFF,0x88111111);
  DEVICE_pinmuxControl(15,0xFFFFFFFF,0x11888888);
  DEVICE_pinmuxControl(16,0xFFFFFFFF,0x11111111);
  DEVICE_pinmuxControl(17,0xFFFFFFFF,0x08111111);
  DEVICE_pinmuxControl(18,0xFFFFFFFF,0x11111111);
  DEVICE_pinmuxControl(19,0xFFFFFFFF,0x00000001);

  return E_PASS;
}

void DEVICE_disable_DSP(void)
{
		/* assert DSP local reset */
	PSC0->MDCTL[15] &= ~0x100;

    /* Turn off DSP */
	DEVICE_LPSCTransition(0, 15, 1, PSC_SWRSTDISABLE);
}


Uint32 DEVICE_SPIInit()
{

  DEVICE_pinmuxControl(7, 0xFFFFF000, 0x11111000);

  return E_PASS;
}


Uint32 DEVICE_TIMER0Init()
{
	return 1;
}

void DEVICE_TIMER0Start(void)
{
}

void DEVICE_TIMER0Stop(void)
{
}

Uint32 DEVICE_TIMER0Status(void)
{
	return 0;
}

Uint32 DEVICE_PLL0Init(Uint32 PLLMult)
{
	register unsigned int PLLM, POSTDIV, PLLDIV3, PLLDIV5, PLLDIV7, CLKMODE, PLL_LOCK_TIME_CNT;

	// Parse arguments
	PLLM = PLLMult;
	POSTDIV = 1;
	PLLDIV3 = 2;
	PLLDIV5 = 5;

	CLKMODE = 0;
	PLLDIV7 = 7;
	PLL_LOCK_TIME_CNT = 2400; 

	// Execute

	// Reset PLL and put in bypass mode
	// Set PLLEN=0 and PLLRST=0
	PLL->PLLCTL &= 0xFFFFFFFE;

	// wait for 4 cycles to allow PLLEN mux switches properly to bypass clock
	UTIL_waitLoop (4);

	// Select the Clock Mode bit 8 as External Clock or On Chip Oscilator
	PLL->PLLCTL &= 0xFFFFFEFF;
	PLL->PLLCTL |= (CLKMODE << 8);

	/* Set PLLENSRC '0',bit 5, PLL Enable(PLLEN) selection is controlled through MMR */
	PLL->PLLCTL &= 0xFFFFFFDF;

	/* PLLCTL.EXTCLKSRC bit 9 should be left at 0 for Primus */
	PLL->PLLCTL &= 0xFFFFFDFF;

	/* Clear PLLRST bit to 0 -Reset the PLL */
	PLL->PLLCTL &= 0xFFFFFFF7;

	/* Disable the PLL output */
	PLL->PLLCTL |= 0x10;

	/* PLL initialization sequence */

	/* Power up the PLL- PWRDN bit set to 0 to bring the PLL out of power down bit */
	PLL->PLLCTL &= 0xFFFFFFFD;

	/* Enable the PLL from Disable Mode PLLDIS bit to 0 - This is step is not required for Primus */
	PLL->PLLCTL &= 0xFFFFFFEF;

	/* Program the required multiplier value in PLLM */
	PLL->PLLM = PLLM;

	/* If desired to scale all the SYSCLK frequencies of a given PLLC, program the POSTDIV ratio */
	PLL->POSTDIV = 0x8000 | POSTDIV;

	/* Check for the GOSTAT bit in PLLSTAT to clear to 0 to indicate that no GO operation is currently in progress */
	while (PLL->PLLSTAT & 0x1 == 1);

	/* Program the RATIO field in PLLDIVx with the desired divide factors. In addition, make sure in this step you leave the PLLDIVx.DxEN bits set so clocks are still enabled (default). */
	PLL->PLLDIV3 = 0x8000 | PLLDIV3;
	PLL->PLLDIV5 = 0x8000 | PLLDIV5;
	PLL->PLLDIV7 = 0x8000 | PLLDIV7;

	/* Set the GOSET bit in PLLCMD to 1 to initiate a new divider transition. */
	PLL->PLLCMD |= 0x1;

	/* Wait for the GOSTAT bit in PLLSTAT to clear to 0 (completion of phase alignment). */
	while (PLL->PLLSTAT & 0x1 == 1);

	/* Set the PLLRST bit in PLLCTL to 1 to bring the PLL out of reset */
	PLL->PLLCTL |= 0x8;

	/* Wait for PLL to lock. See PLL spec for PLL lock time */
	UTIL_waitLoop (PLL_LOCK_TIME_CNT);

	/* Set the PLLEN bit in PLLCTL to 1 to remove the PLL from bypass mode */
	PLL->PLLCTL |= 0x1;

	SYSTEM->KICKR[0] = 0x83e70b13;  // Kick0 register + data (unlock)
	SYSTEM->KICKR[1] = 0x95a4f1e0;  // Kick1 register + data (unlock)

	SYSTEM->CFGCHIP[3] |= 0x4;       // Enable 4.5 divider PLL
	SYSTEM->CFGCHIP[3] |= 0x3;       // Select 4.5 divider for EMIFA / EMIFB clock source

	SYSTEM->KICKR[0] = 0x00000000;  // Kick0 register + data (lock)
	SYSTEM->KICKR[1] = 0x00000000;  // Kick1 register + data (lock)

	return E_PASS;
}

Uint32 DEVICE_SDRAMInit()
{
	DEVICE_pinmuxControl(0, 0xFFFFFF00, 0x11111100);
	DEVICE_pinmuxControl(1, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(2, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(3, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(4, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(5, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(6, 0xFFFFFFFF, 0x11111111);
	DEVICE_pinmuxControl(7, 0x00000FFF, 0x00000111);

	// Configure
	DDR->SDCR 	= 0x00018421;
	DDR->SDTIMR = 0x10912A08;
	DDR->SDTIMR2 = 0x70090005;
	DDR->SDRCR	= 0x000003FA;

	    // ISSI IS42S16160B-6BL SDRAM, 2 x 16M x 16 (32-bit data path), 133MHz
    DDR->SDCR = 0          // SDRAM Bank Config Register
        |( 1 << 15)         // Unlock timing registers
        |( 2 << 9 )         // CAS latency is 2
        |( 2 << 4 )         // 4 bank SDRAM devices
        |( 1 << 0 );        // 512-word pages requiring 9 column address bits

    DDR->SDRCR = 0         // SDRAM Refresh Control Register
        |( 0 << 31)         // Low power mode disabled
        |( 0 << 30)         // MCLK stoping disabled
        |( 0 << 23)         // Selects self refresh instead of power down
        |( 1040 <<0);       // Refresh rate = 7812.5ns / 7.5ns

    DDR->SDTIMR = 0       // SDRAM Timing Register 1
        |( 25 << 25)        // (67.5ns / 7.55ns) - 1 = TRFC  @ 133MHz
        |( 2 << 22 )        // (20ns / 7.5ns) - 1 =TRP
        |( 2 << 19 )        // (20ns / 7.5ns) - 1 = TRCD
        |( 1 << 16 )        // (14ns / 7.5ns) - 1 = TWR
        |( 5 << 11 )        // (45ns / 7.5ns) - 1 = TRAS
        |( 8 <<  6 )        // (67.5ns / 7.5ns) - 1 = TRC
        |( 2 <<  3 );       // *(((4 * 14ns) + (2 * 7.5ns)) / (4 * 7.5ns)) -1. = TRRD
                            // but it says to use this formula if 8 banks but only 4 are used here.
                            // and SDCFG1 register only suports upto 4 banks.

    DDR->SDTIMR2 = 0       // SDRAM Timing Register 2
        |( 14<< 27)         // not sure how they got this number. the datasheet says value should be
                            // "Maximum number of refresh_rate intervals from Activate to Precharge command"
                            // but has no equation. TRASMAX is 120k.
        |( 9 << 16)         // ( 70 / 7.5) - 1
        |( 5 << 0 );        // ( 45 / 7.5 ) - 1

    DDR->SDCR = 0          // SDRAM Bank Config Register
	    |( 1 << 16)
        |( 0 << 15)         // Unlock timing registers
        |( 2 << 9 )         // CAS latency is 2
        |( 2 << 4 )         // 4 bank SDRAM devices
        |( 1 << 0 );        // 512-word pages requiring 9 column address bits

	return E_PASS;
}

/************************************************************
* Local Function Definitions                                *
************************************************************/


/***********************************************************
* End file                                                 *
***********************************************************/

/* --------------------------------------------------------------------------
    HISTORY
        v1.0 completion 							 						      
            Daniel Allred  -  1-Nov-2007
 ----------------------------------------------------------------------------- */



